# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Terms updater."""

import copy
import re
from collections import OrderedDict

from invenio_access.permissions import system_identity
from invenio_db import db
from invenio_pidstore.models import PersistentIdentifier
from invenio_rdm_records.records import RDMDraft, RDMRecord
from invenio_records_resources.proxies import current_service_registry
from invenio_records_resources.services.uow import RecordCommitOp
from invenio_search.engine import search
from sqlalchemy import delete, select

from .keeptrace import KeepTrace

bulk = search.helpers.bulk


def filter_ops_by_type(ops_data, _type):
    """Filter ops_data by _type."""
    return [op for op in ops_data if op.get("type") == _type]


def find_idx_subject_dict(subjects, id_):
    """Return (idx, subject_dict) of subject with `id_` in `subjects`.

    Return (-1, {}) if not found.
    """
    return next(
        ((i, s) for i, s in enumerate(subjects) if s.get("id") == id_),
        (-1, {})
    )


def op_replace(subjects, op_data):
    """Replace subject in-place.

    op_data:

    ```python
        {
            "type": "replace",
            "scheme": "...",
            "id": "...",
            "new_id": "...",
            "new_subject": "..."
        }
    ```
    """
    idx, subject_dict = find_idx_subject_dict(subjects, op_data["id"])
    if idx == -1:
        return False
    new_subject_dict = {
        # **subject_dict,  # probably not needed
        "id": op_data["new_id"],
    }
    subjects[idx] = new_subject_dict
    return True


def op_remove(subjects, op_data):
    """Remove subject in-place.

    op_data:

    ```python
        {
            "type": "remove",
            "scheme": "...",
            "id": "...",
        }
    ```
    """
    idx, subject_dict = find_idx_subject_dict(subjects, op_data["id"])
    if idx == -1:
        return False
    subjects.pop(idx)
    return True


def op_rename(subjects, op_data):
    """Fake rename subject in-place.

    Renaming is not done at the record DB level since a record's subjects
    are dereferenced dynamically. We do need to know if the record would
    see a rename in one of its subjects, so that we can flag it for
    keeping track of former subject + logging.

    The passed subjects should also be the original subjects of the record.
    Otherwise, a replace X for Y, followed by rename of Y's label could
    leave a trace on the record as though the record used to have Y's
    previous label.
    """
    idx, subject_dict = find_idx_subject_dict(subjects, op_data["id"])
    if idx == -1:
        return False
    # no action as mentioned
    return True


def apply_op_data_change(op_data, orig_subjects, record):
    """Apply `op_data` change on `record`."""
    subjects = record["metadata"]["subjects"]
    applied = False
    if op_data.get("type") == "replace":
        applied = op_replace(subjects, op_data)
    elif op_data.get("type") == "remove":
        applied = op_remove(subjects, op_data)
    elif op_data.get("type") == "rename":
        applied = op_rename(orig_subjects, op_data)
    return applied


def deduplicate_subjects(subjects):
    """Return a deduplicated (order preserved) list of subjects.

    :param subjects: list[dict]
    :returns: list[dict]
    """
    def _tuplify(d):
        """Transforms a subject dict into a tuple with value stripped.

        This is needed because dicts are unhashable.
        """
        key = "id" if "id" in d else "subject"  # only 2 possibilities
        value = d.get(key).strip()
        return (key, value)

    result = [_tuplify(d) for d in subjects]
    result = OrderedDict.fromkeys(result)  # rm duplicates AND keeps order
    result = [dict([t]) for t in result.keys()]  # back to subjects
    return result


def update_rdm_record(record, ops_data, logger, keep_trace):
    """Apply changes to record's subjects."""
    orig_subjects = copy.deepcopy(record["metadata"]["subjects"])
    for op_data in ops_data:
        applied = apply_op_data_change(
            op_data,
            orig_subjects,
            record,
        )

        if applied:
            if keep_trace.should_trace(op_data):
                keep_trace.trace(record, op_data["subject"])
            logger.log(record.pid.pid_value, delta=op_data)

    # Make sure subjects are deduplicated
    record["metadata"]["subjects"] = deduplicate_subjects(
        record["metadata"]["subjects"]
    )

    # All side-effects
    # ---
    records_service = current_service_registry.get("records")
    # According to other code in InvenioRDM, the same indexer is used for
    # records and drafts
    commit_op = RecordCommitOp(record, records_service.indexer)
    # The following on_register, on_commit don't use the uow object
    # so passing None is fine
    fake_uow = None
    try:
        commit_op.on_register(fake_uow)  # commits to DB
        commit_op.on_commit(fake_uow)  # reindexes in index
    except Exception as e:
        msg = re.sub(r"\s+", " ", str(e))
        logger.log(record.pid.pid_value, error=msg)
    finally:
        logger.flush()


def get_records_to_update(ops_data, data_cls):
    """Return data-layer records to update."""

    def get_targeted_ids(ops_data):
        return [
            op["id"] for op in ops_data
            if op.get("type") in ["replace", "remove", "rename"]
        ]

    def has_at_least_1_subject_targeted(record_data_db, ids):
        """Return True if `record_data_db` has at least 1 subject in `ids`.

        Because of cases where there are 100K+ subjects, we can't construct
        queries filtering at the DB level. We do the filtering in memory,
        on batches of records. We keep loading simple at the cost of
        performance (but this is fine since this operation is in the
        background anyway).
        """
        # May be None in some Drafts
        if not record_data_db:
            return False
        subjects = record_data_db.get("metadata", {}).get("subjects", [])
        return any(s.get("id") in ids for s in subjects)

    targeted_ids = frozenset(get_targeted_ids(ops_data))

    if not targeted_ids:
        return []

    stmt = (
        select(data_cls.model_cls)
        .execution_options(yield_per=200)  # could be made adjustable
    )

    result = (
        data_cls(obj.data, model=obj) for obj
        in db.session.scalars(stmt)
        if has_at_least_1_subject_targeted(obj.data, targeted_ids)
    )

    return result


class SubjectDeltaUpdater:
    """Translates delta operations into actual changes."""

    def __init__(self, ops_data, logger, keep_trace):
        """Constructor."""
        self._ops_data = ops_data
        self._logger = logger
        self._keep_trace = keep_trace

    def update(self):
        """Execute changes."""
        self._add_rdm_subjects()
        self._rename_rdm_subjects()

        self._update_rdm_records()

        self._remove_rdm_subjects()

    def _add_rdm_subjects(self):
        """Add to the Subjects entries."""
        service = current_service_registry.get("subjects")
        add_ops = filter_ops_by_type(self._ops_data, "add")
        for op in add_ops:
            service.create(
                system_identity,
                {
                    "id": op["id"],
                    "scheme": op["scheme"],
                    "subject": op["subject"],
                }
            )

    def _rename_rdm_subjects(self):
        """Rename subjects in the Subjects entries."""
        service = current_service_registry.get("subjects")
        rename_ops = filter_ops_by_type(self._ops_data, "rename")
        for op in rename_ops:
            service.update(
                system_identity,
                op["id"],
                {
                    "id": op["id"],
                    "scheme": op["scheme"],
                    "subject": op["new_subject"]
                }
            )

    def _update_rdm_records(self):
        """Execute operations (replace/remove/rename) on RDM records."""
        entries = get_records_to_update(self._ops_data, data_cls=RDMRecord)
        for record in entries:
            update_rdm_record(
                record,
                ops_data=self._ops_data,
                logger=self._logger,
                keep_trace=self._keep_trace
            )

        # Don't keep trace for drafts
        entries = get_records_to_update(self._ops_data, data_cls=RDMDraft)
        for draft in entries:
            update_rdm_record(
                draft,
                ops_data=self._ops_data,
                logger=self._logger,
                keep_trace=KeepTrace(None, None)  # noop KeepTrace
            )

    def _remove_rdm_subjects(self):
        """Remove subjects from the Subjects entries.

        We have to resort to low-level commands because the high-level ones
        are not made for bulk operations. We've checked the implications
        and we should be fine (at least at time of writing).
        """
        ids_for_removal = [
            op["id"] for op in self._ops_data
            if op.get("type") in ["remove", "replace"]
        ]
        service = current_service_registry.get("subjects")

        model_cls = service.record_cls.model_cls
        size_of_batch = 200  # Maybe TODO: make configurable
        for offset in range(0, len(ids_for_removal), size_of_batch):
            batch = ids_for_removal[offset:offset + size_of_batch]

            # Delete from database
            # ===
            # Get ids of model class
            # The ids_for_removal internally correspond to
            # pids, so they need to be dereferenced first
            stmt_to_select_ids = (
                select(model_cls.id)
                .where(model_cls.id == PersistentIdentifier.object_uuid)
                .where(PersistentIdentifier.pid_type == "sub")
                .where(PersistentIdentifier.pid_value.in_(batch))
            )
            ids_of_model_cls = list(db.session.scalars(stmt_to_select_ids))

            # Delete subject records
            stmt_to_delete_subjects = (
                delete(model_cls)
                .where(model_cls.id.in_(ids_of_model_cls))
                # ORM session synchronization has to be specified when deleting
                # Here we skip synchronization since not needed
                .execution_options(synchronize_session=False)
            )
            db.session.execute(stmt_to_delete_subjects)

            # Delete backing subject PID
            stmt_to_delete_pids = (
                delete(PersistentIdentifier)
                .where(PersistentIdentifier.pid_type == "sub")
                .where(PersistentIdentifier.pid_value.in_(batch))
                .execution_options(synchronize_session=False)
            )
            db.session.execute(stmt_to_delete_pids)

            db.session.commit()

            # Delete from document engine
            # ===
            alias_of_index = service.record_cls.index.search_alias
            bulk(
                service.indexer.client,
                (
                    {
                        "_op_type": "delete",
                        "_index": alias_of_index,
                        "_id": id_
                    }
                    for id_ in ids_of_model_cls
                ),
            )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test MeSH-related download functionality."""

from collections import namedtuple
from contextlib import contextmanager
from pathlib import Path
from unittest import mock

from galter_subjects_utils.contrib.mesh.downloader import MeSHDownloader

# Helpers


@contextmanager
def fake_request_context(url, stream):
    """A faked requests.get context result."""
    filename = url.rsplit("/", 1)[-1]
    fp = Path(__file__).parent / f"data/fake_{filename}"

    FakeRequestContext = namedtuple("FakeRequestContext", ["raw"])

    with open(fp, "rb") as f:
        yield FakeRequestContext(raw=f)


# Tests


@mock.patch('galter_subjects_utils.downloader.requests.get')
def test_downloader(patched_get, tmp_path):
    downloads_dir = tmp_path / "downloads"
    downloads_dir.mkdir()
    # patch requests.get to return files
    patched_get.side_effect = fake_request_context

    # Testing the current_year variant seemed like it would lead to a test
    # mirror, so it has been tested manually instead. This really just tests
    # that we have the right name, interfaces and final files written to disk
    # in right location.
    downloader = MeSHDownloader(
        year="2023",
        directory=downloads_dir
    )
    downloader.download()

    patched_get.assert_called()
    assert (downloads_dir / "d2023.bin").exists()
    assert (downloads_dir / "q2023.bin").exists()
    assert (downloads_dir / "meshnew2023.txt").exists()
    assert (downloads_dir / "replace2023.txt").exists()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Download MeSH files."""

from datetime import date

from galter_subjects_utils.downloader import download_file


class MeSHDownloader:
    """Download MeSH files."""

    terms_filename_tmpl = "d{year}.bin"
    qualifiers_filename_tmpl = "q{year}.bin"
    new_filename_tmpl = "meshnew{year}.txt"
    replace_filename_tmpl = "replace{year}.txt"

    def __init__(self, directory, year, cache=False):
        """Constructor."""
        self.directory = directory
        self.year = str(year)
        self.cache = cache

        self.base_url = "https://nlmpubs.nlm.nih.gov/projects/mesh"

    def file_urls(self):
        """List of files to download."""
        current_year = str(date.today().year)
        url_parent_segment = "MESH_FILES" if self.year == current_year else self.year  # noqa
        base_url = self.base_url + "/" + url_parent_segment

        urls = []

        # asciimesh files
        base_url_asciimesh = base_url + "/" + "asciimesh"
        urls += [
            base_url_asciimesh + "/" + fn_tmpl.format(year=self.year)
            for fn_tmpl in [
                self.terms_filename_tmpl,
                self.qualifiers_filename_tmpl
            ]
        ]

        # newterms files
        base_url_newterms = base_url + "/" + "newterms"
        urls += [
            base_url_newterms + "/" + fn_tmpl.format(year=self.year)
            for fn_tmpl in [
                self.new_filename_tmpl,
                self.replace_filename_tmpl
            ]
        ]

        return urls

    def download(self):
        """Download MeSH files of interest."""
        # files to download
        urls = self.file_urls()

        for url in urls:
            dest_path = self.directory / url.rsplit("/", 1)[-1]
            if self.cache and dest_path.exists():
                continue
            download_file(url, dest_path)

    @property
    def terms_filepath(self):
        """Downloaded terms filepath."""
        return self.directory / self.terms_filename_tmpl.format(year=self.year)

    @property
    def qualifiers_filepath(self):
        """Downloaded qualifiers filepath."""
        return self.directory / self.qualifiers_filename_tmpl.format(year=self.year)  # noqa

    @property
    def new_filepath(self):
        """Downloaded meshnew filepath."""
        return self.directory / self.new_filename_tmpl.format(year=self.year)

    @property
    def replace_filepath(self):
        """Downloaded replace filepath."""
        return self.directory / self.replace_filename_tmpl.format(year=self.year)  # noqa



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Conftest."""

import pytest
from invenio_access.permissions import system_identity
from invenio_app.factory import create_app as _create_app
from invenio_records_resources.proxies import current_service_registry


@pytest.fixture(scope='module')
def app_config(app_config):
    """Override pytest-invenio app_config fixture.

    For test purposes we need to explicitly set these configuration variables
    above any other module's config.py potential clashes.
    """
    app_config.update({
        # Variable not used. We set it to silent warnings
        'JSONSCHEMAS_HOST': 'not-used',
        # Disable DATACITE.
        'RDM_RECORDS_DOI_DATACITE_ENABLED': False,
        'RECORDS_REFRESOLVER_CLS': (
            "invenio_records.resolver.InvenioRefResolver"
        ),
        'RECORDS_REFRESOLVER_STORE': (
            "invenio_jsonschemas.proxies.current_refresolver_store"
        ),
        'MAIL_DEFAULT_SENDER': ('Prism', 'no-reply@localhost'),
        # Uncomment to investigate SQL queries
        # 'SQLALCHEMY_ECHO': True,
    })

    return app_config


@pytest.fixture(scope='module')
def create_app():
    """Create app fixture for UI+API app."""
    return _create_app


# Plethora of required pre-existing data to create a record


def create_vocabulary_type(id_, pid):
    """Create vocabulary type."""
    vocabulary_service = current_service_registry.get("vocabularies")
    return vocabulary_service.create_type(system_identity, id_, pid)


@pytest.fixture(scope="module")
def resourcetypes_type(app):
    """Resource type vocabulary type."""
    return create_vocabulary_type("resourcetypes", "rsrct")


@pytest.fixture(scope="module")
def resourcetypes(app, resourcetypes_type):
    """Resource type vocabulary record."""
    vocabulary_service = current_service_registry.get("vocabularies")
    vocabulary_service.create(
        system_identity,
        {  # create parent resource type
            "id": "text",
            "title": {
                "en": "Text Resources"
            },
            "type": "resourcetypes"
        }
    )
    program = vocabulary_service.create(
        system_identity,
        {
            "icon": "file alternate",
            "id": "text-program",
            "props": {
                "coar": "text",
                "csl": "",
                "datacite_general": "Text",
                "datacite_type": "Program",
                "eurepo": "info:eu-repo/semantics/other",
                "openaire_resourceType": "",
                "openaire_type": "Text",
                "schema.org": "https://schema.org/TextDigitalDocument",
                "subtype": "text-program",
                "type": "text",
            },
            "title": {
                "en": "Program"
            },
            "tags": ["depositable", "linkable"],
            "type": "resourcetypes"
        }
    )
    return [program]


@pytest.fixture(scope="module")
def running_app(app, location, resourcetypes):
    """Running app."""
    return app


@pytest.fixture(scope="module")
def minimal_record_input():
    """Record input dict."""
    return {
        "pids": {},
        "access": {
            "record": "public",
            "files": "public",
        },
        "files": {
            "enabled": False,  # Most tests don't care about files
        },
        "metadata": {
            "creators": [
                {
                    "person_or_org": {
                        "family_name": "Brown",
                        "given_name": "Troy",
                        "type": "personal",
                    }
                },
                {
                    "person_or_org": {
                        "name": "Troy Inc.",
                        "type": "organizational",
                    },
                },
            ],
            "publication_date": "2020-06-01",
            # because DATACITE_ENABLED is True, this field is required
            "publisher": "Acme Inc",
            "resource_type": {"id": "text-program"},
            "title": "A Romans story",
        },
    }


@pytest.fixture(scope="module")
def create_draft_data(running_app, minimal_record_input):
    """Draft data-layer fixture."""
    records_service = current_service_registry.get("records")

    def _create_draft_data(
        identity=None,
        record_input=minimal_record_input,
    ):
        """Create a data-layer draft record."""
        identity = identity or system_identity
        draft_data = records_service.create(identity, record_input)._record
        return draft_data

    return _create_draft_data


@pytest.fixture(scope="module")
def create_record_data(create_draft_data, minimal_record_input):
    """Record data-layer fixture."""
    record_service = current_service_registry.get("records")

    def _create_record_data(
        identity=None,
        record_input=minimal_record_input,
    ):
        """Create a data-layer record."""
        identity = identity or system_identity
        draft_data = create_draft_data(identity, record_input)
        record_result = record_service.publish(
            identity, draft_data.pid.pid_value
        )
        return record_result._record

    return _create_record_data


@pytest.fixture(scope="function")
def create_record_data_fn(create_record_data, db, search):
    """Function scoped record creation fixture."""

    def _create_record_data(identity, record_input):
        """Create a data-layer record."""
        return create_record_data(identity, record_input)

    return _create_record_data


@pytest.fixture(scope="function")
def minimal_subject_input():
    """Minimal subject input dict."""
    return {
        "id": "http://example.org/foo/0",
        "scheme": "foo",
        "subject": "0",
    }


@pytest.fixture(scope="function")
def create_subject_data(running_app, db, search, minimal_subject_input):
    """Create data-layer subjects."""
    subjects_service = current_service_registry.get("subjects")

    def _create_subject_data(
        identity=None,
        input_=None
    ):
        """Create a data-layer subject."""
        identity = identity or system_identity
        input_ = input_ or minimal_subject_input
        result = subjects_service.create(
            identity,
            input_
        )
        return result._record

    return _create_subject_data



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test general reader functionality."""


from galter_subjects_utils.reader import get_rdm_subjects, mapping_by


def test_mapping_by():
    iterable = [
        {
            "RECTYPE": "D",
            "MH": "Volvox",
            "AQ": [
                "CH", "CL", "CY", "DE", "EN", "GD", "GE", "IM", "IP", "ME",
                "MI", "PH", "PS", "PY", "RE", "UL", "VI"
            ],
            "MN": "B01.875.150.950",
            "MH_TH": "NLM (1998)",
            "ST": "T002",
            "RN": "txid3066",
            "PI": "Algae, Green (1970-2003)",
            "MS": "A genus of GREEN ALGAE in the family Volvocaceae...",
            "PM": "2004; see ALGAE, GREEN 1998-2003",
            "HN": "2004; use ALGAE, GREEN 1998-2003",
            "MR": "20210630",
            "DA": "20030709",
            "DC": "1",
            "DX": "20040101",
            "UI": "D044446",
        },
        {
            "RECTYPE": "D",
            "MH": "American Native Continental Ancestry Group",
            "AQ": ["CL", "ED", "EH", "GE", "HI", "LJ", "PX", "SN"],
            "MN": "M01.686.508.150",
            "MH_TH": "NLM (2004)",
            "ST": "T098",
            "MS": "Individuals whose ancestral origins are in the ...",
            "PM": "2004",
            "HN": "2004",
            "DA": "20030709",
            "DC": "1",
            "DX": "20040101",
            "UI": "D044467",
        }
    ]

    mapping = mapping_by(iterable, by="UI", keys=["UI", "MH", "AQ"])

    expected = {
        "D044446": {
            "UI": "D044446",
            "MH": "Volvox",
            "AQ": [
                "CH", "CL", "CY", "DE", "EN", "GD", "GE", "IM", "IP", "ME",
                "MI", "PH", "PS", "PY", "RE", "UL", "VI"
            ],
        },
        "D044467": {
            "UI": "D044467",
            "MH": "American Native Continental Ancestry Group",
            "AQ": ["CL", "ED", "EH", "GE", "HI", "LJ", "PX", "SN"],
        }
    }
    assert expected == mapping


def test_get_rdm_subjects(running_app, create_subject_data):
    subjects = [
        create_subject_data(
            input_={
                "id": "http://example.org/foo/0",
                "scheme": "foo",
                "subject": "0",
            }
        ),
        create_subject_data(
            input_={
                "id": "http://example.org/foo/1",
                "scheme": "foo",
                "subject": "1",
            },
        ),
        create_subject_data(
            input_={
                "id": "http://example.org/foo/2",
                "scheme": "foo",
                "subject": "2",
            },
        ),
        create_subject_data(
            input_={
                "id": "http://example.org/bar/0",
                "scheme": "bar",
                "subject": "0",
            },
        )
    ]

    subjects = [s for s in get_rdm_subjects(scheme="foo")]
    assert 3 == len(subjects)
    assert {"0", "1", "2"} == {e["subject"] for e in subjects}



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""galter-subjects-utils for InvenioRDM."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Adapters to convert between different outputs and inputs."""

from .converter import Subject


def generate_replacements(replace_iter):
    """Converts a replace iterable into a replacement dict.

    :param replace_iter: Iterable[ReplaceEntry]

    ReplaceEntry:

    {
        "MH OLD": "...",
        "MH NEW": "...",
        "delete": "...",
        "status": "...",
    }
    """
    return {
        e["MH OLD"]: e["MH NEW"] for e in replace_iter if e["delete"]
    }


def converted_to_subjects(rdm_iter):
    """Adapts an RDMSubject ("Converted") iterable into a Subject iterable.

    :param rdm_iter: Iterable[RDMSubject]

    RDMSubject:

    {
        "id": "https://id.nlm.nih.gov/mesh/...",
        "scheme": "...",
        "subject": "..."
    }

    :return: Iterable[Subject]

    Subject:

        id (minus the https://id.nlm.nih.gov/mesh/ part)
        label
    """
    prefix = "https://id.nlm.nih.gov/mesh/"
    len_prefix = len(prefix)
    return (
        Subject(
            id=e["id"][len_prefix:] if e["id"].startswith(prefix) else e["id"],
            label=e["subject"]
        ) for e in rdm_iter
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Generic conversion functionality."""


class LCSHRDMConverter:
    """Convert LCSH term into RDM subjects dict."""

    def __init__(self, topics):
        """Constructor.

        :param topics: LCSH Topics iterable
        :type topics: iterable[dict]
        """
        self.topics = topics

    def __iter__(self):
        """Iterate over terms."""
        for topic in self.topics:
            entry = self.extract_entry(topic)
            if entry:
                yield entry

    def extract_entry(self, topic):
        """Extract relevant dict from LCSH skos json-ld dict."""
        # Find id and label
        id_suffix = topic["@id"]
        info_entry = next(
            (
                e for e in topic["@graph"]
                if (
                    e.get("@id").endswith(id_suffix) and
                    e.get("@type") == "skos:Concept"
                )
            ),
            None
        )

        if not info_entry:
            return {}

        entry = {
            "id": info_entry["@id"].replace("http://", "https://"),
            "scheme": "LCSH"
        }
        pref_label = info_entry["skos:prefLabel"]
        if isinstance(pref_label, dict):
            subject = pref_label["@value"]
        elif isinstance(pref_label, list):
            subject = pref_label[-1]["@value"]
        else:
            raise Exception("invalid skos:prefLabel")
        entry["subject"] = subject

        return entry



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Convert MeSH data to InvenioRDM/ops."""

from dataclasses import dataclass


def generate_id(identifier):
    """Generate URI id."""
    return f"https://id.nlm.nih.gov/mesh/{identifier}"


class MeSHRDMConverter:
    """Convert MeSH term into RDM subject dict."""

    def __init__(self, topics, qualifiers_mapping=None):
        """Constructor.

        :param topics: MeSH Topics iterable
        :type topics: iterable[dict]
        :param qualifiers: MeSH qualifiers iterable, defaults to None
        :type qualifiers: iterable[dict], optional
        """
        self.topics = topics
        self.qualifiers_mapping = qualifiers_mapping

    def __iter__(self):
        """Iterate over converted entries.

        DEPRECATED.
        """
        yield from self.convert()

    def convert(self):
        """Iterator over converted entries."""
        for topic in self.topics:
            yield {
                "id": generate_id(topic['UI']),
                "scheme": "MeSH",
                "subject": topic['MH']
            }

            if not self.qualifiers_mapping:
                continue

            for qid in topic.get("AQ", []):
                qualifier = self.qualifiers_mapping.get(qid)

                if not qualifier:
                    continue

                yield {
                    "id": generate_id(topic['UI'] + qualifier['UI']),
                    "scheme": "MeSH",
                    "subject": topic['MH'] + "/" + qualifier['SH']
                }


@dataclass
class Subject:
    """Minimalist subject data."""

    id: str
    label: str


@dataclass
class Analysis:
    """Holds results of analyzing a src subject."""

    id: str
    seen: bool = False  # present in dst subjects
    relabelled: str = ""  # new label
    replaced: str = ""  # id of replacing subject


class MeSHSubjectDeltasConverter:
    """Converter between subjects in DB + new subject file to operations."""

    def __init__(self, src_subjects, dst_subjects, replacements):
        """Constructor.

        :param src_subjects: MeSH subjects found in the instance
        :type src_subjects: Iterable[Subject]
        :param dst_subjects: new subjects reader
        :type dst_subjects: Iterable[Subject]
        :param replacements: mapping of src labels to dst labels
        :type replacements: dict[str, str]
        """
        self.src_subjects = src_subjects
        self.dst_subjects = dst_subjects
        self.replacements = replacements

        # Will be filled out by _analyze_src(), _analyze_dst() and
        # _analyze_replace()
        self._id_to_analysis = {}
        self._label_to_analysis = {}
        self._additions = []

        self.scheme = "MeSH"

    def convert(self):
        """Generate operations."""
        self._analyze_src()
        self._analyze_dst()
        self._analyze_replace()
        return self._generate_ops()

    def _analyze_src(self):
        """Build internal analysis.

        The self.src_subjects iterable is completely iterated and kept in
        memory (in constituent parts). This allows us quick access and frees
        us from having to load all the self.dst_subjects as well.
        """
        for subject in self.src_subjects:
            analysis = Analysis(id=subject.id)
            self._id_to_analysis[subject.id] = analysis
            self._label_to_analysis[subject.label] = analysis

    def _analyze_dst(self):
        """Analyze through all dst subjects."""
        for dst_subject in self.dst_subjects:
            # A dst subject can be
            # added/new ?
            if dst_subject.id not in self._id_to_analysis:
                self._additions.append(dst_subject)
            # relabelled ?
            elif dst_subject.label not in self._label_to_analysis:
                analysis = self._id_to_analysis[dst_subject.id]
                analysis.seen = True
                analysis.relabelled = dst_subject.label
            # identical ?
            else:
                analysis = self._id_to_analysis[dst_subject.id]
                analysis.seen = True

    def _analyze_replace(self):
        """Determine which subjects have been replaced."""

        def find_replacement_label(label):
            """Return replacement label or None."""
            new_label = self.replacements.get(label)
            if new_label:
                return new_label

            # maybe label is qualified, so check if root is replaced
            root_label, slash, qualifier = label.rpartition("/")
            new_root_label = self.replacements.get(root_label)
            if not new_root_label:
                return None
            new_label = new_root_label + slash + qualifier
            return new_label

        def find_replacement_id(label):
            """Return replacement id or None if not replaced."""
            replacement_label = find_replacement_label(label)
            if not replacement_label:
                return None

            # is replacement_label for existing and kept subject?
            replacement_analysis = self._label_to_analysis.get(replacement_label)  # noqa
            if replacement_analysis and replacement_analysis.seen:
                return replacement_analysis.id

            # is replacement_label for newly added subject?
            replacement_subject = next(
                (s for s in self._additions if s.label == replacement_label),
                None
            )
            if replacement_subject:
                return replacement_subject.id

            # is replacement_label for relabelled existing and kept subject?
            replacement_analysis = next(
                (
                    a for a in self._label_to_analysis.values()
                    if a.relabelled == replacement_label
                ),
                None
            )
            if replacement_analysis:
                return replacement_analysis.id

            return None

        for label, analysis in self._label_to_analysis.items():
            if analysis.seen:
                continue

            replacement_id = find_replacement_id(label)
            if replacement_id:
                analysis.replaced = replacement_id

    def _generate_ops(self):
        """Generate delta operations."""
        result = []

        # Additions
        result.extend(
            {
                "type": "add",
                "scheme": self.scheme,
                "id": generate_id(a.id),
                "subject": a.label
            }
            for a in self._additions
        )

        # Renames
        result.extend(
            {
                "type": "rename",
                "scheme": self.scheme,
                "id": generate_id(analysis.id),
                "subject": label,
                "new_subject": analysis.relabelled
            }
            for label, analysis in self._label_to_analysis.items()
            if analysis.relabelled
        )

        # Replaces
        result.extend(
            {
                "type": "replace",
                "scheme": self.scheme,
                "id": generate_id(analysis.id),
                "subject": label,
                "new_id": generate_id(analysis.replaced)
            }
            for label, analysis in self._label_to_analysis.items()
            if analysis.replaced
        )

        # Removes
        result.extend(
            {
                "type": "remove",
                "scheme": self.scheme,
                "id": generate_id(analysis.id),
                "subject": label
            }
            for label, analysis in self._label_to_analysis.items()
            if not analysis.seen and not analysis.replaced
        )

        return result



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2023 Northwestern University.
#
# invenio-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

from collections import namedtuple
from contextlib import contextmanager
from pathlib import Path
from unittest import mock

from galter_subjects_utils.converter import LCSHRDMConverter
from galter_subjects_utils.downloader import LCSHDownloader
from galter_subjects_utils.reader import read_jsonl
from galter_subjects_utils.writer import write_jsonl

# Helpers


@contextmanager
def fake_request_context(url, stream):
    fp = Path(__file__).parent / "data/fake_lcsh.skosrdf.jsonld.gz"
    FakeRequestContext = namedtuple("FakeRequestContext", ["raw"])
    with open(fp, "rb") as f:
        yield FakeRequestContext(raw=f)


# Tests


@mock.patch('galter_subjects_utils.downloader.requests.get')
def test_downloader(patched_get, tmp_path):
    # patch requests.get to return files
    patched_get.side_effect = fake_request_context
    downloads_dir = tmp_path / "downloads"
    downloads_dir.mkdir()

    # No cache by default
    downloader = LCSHDownloader(directory=downloads_dir)
    downloader.download()

    patched_get.assert_called()
    assert (
        downloads_dir / "subjects.skosrdf.jsonld.gz" ==
        downloader.downloaded_filepath
    )
    assert (
        downloads_dir / "subjects.skosrdf.jsonld" ==
        downloader.extracted_filepath
    )

    # With cache
    patched_get.reset_mock()
    downloader = LCSHDownloader(directory=downloads_dir, cache=True)
    downloader.download()

    patched_get.assert_not_called()
    assert (
        downloads_dir / "subjects.skosrdf.jsonld.gz" ==
        downloader.downloaded_filepath
    )
    assert (
        downloads_dir / "subjects.skosrdf.jsonld" ==
        downloader.extracted_filepath
    )


def test_converter():
    # File is setup to test
    # - regular entries
    # - deprecated entry that is ignored
    # - entry with multiple labels (take the last)
    filepath = Path(__file__).parent / "data" / "fake_lcsh.skosrdf.jsonld"
    topics = list(read_jsonl(filepath))
    converter = LCSHRDMConverter(topics)

    objects = [o for o in converter]

    expected = [
        {
            "id": 'https://id.loc.gov/authorities/subjects/sh00000011',
            "scheme": "LCSH",
            "subject": "ActionScript (Computer program language)"
        },
        {
            "id": 'https://id.loc.gov/authorities/subjects/sh00000014',
            "scheme": "LCSH",
            "subject": "Tacos"
        },
        {
            "id": 'https://id.loc.gov/authorities/subjects/sh90000997',
            "scheme": "LCSH",
            "subject": "Rooms"
        }
    ]
    assert expected == objects


def test_write():
    filepath = Path(__file__).parent / "test_subjects_lcsh.jsonl"
    entries = [
        {
            "id": 'https://id.loc.gov/authorities/subjects/sh00000011',
            "scheme": "LCSH",
            "subject": "ActionScript (Computer program language)"
        },
        {
            "id": 'https://id.loc.gov/authorities/subjects/sh00000014',
            "scheme": "LCSH",
            "subject": "Tacos"
        },
        {
            "id": 'https://id.loc.gov/authorities/subjects/sh00000275',
            "scheme": "LCSH",
            "subject": "Shell Lake (Wis. : Lake)"
        },
        {
            "id": 'https://id.loc.gov/authorities/subjects/sh00008126',
            "scheme": "LCSH",
            "subject": "Half-Circle \"V\" Ranch (Wyo.)"
        },
    ]

    write_jsonl(entries, filepath)

    read_entries = list(read_jsonl(filepath))
    assert entries == read_entries

    filepath.unlink(missing_ok=True)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Read MeSH files."""

import re

# Utilities


def topic_filter(record):
    """Filters for topical terms."""
    return record.get("DC") == "1"


# Readers


class MeSH:
    """MeSH term extractor.

    DEPRECATED. Will be removed in following versions.
    """

    @classmethod
    def load(cls, filepath, filter=lambda r: True):
        """Iterate over selected MeSH terms."""
        with open(filepath, 'r') as f:
            record = {}

            for line in f.readlines():
                if "=" not in line:
                    continue

                key, value = [p.strip() for p in line.split("=", maxsplit=1)]
                cls.update_record(record, key, value)

                # Assumes always ends with UI
                if key == "UI":
                    if filter(record):
                        yield record
                    record = {}

    @classmethod
    def update_record(cls, record, key, value):
        """Update the value of a key in the MeSH record.

        We deal with special cases of interest here.
        """
        if key == "AQ":
            record[key] = [q.strip() for q in value.split()]
        else:
            record[key] = value


class MeSHReader:
    """MeSH Reader."""

    def __init__(self, filepath, filter=lambda r: True):
        """Constructor."""
        self._filepath = filepath
        self._filter = filter

    def __iter__(self):
        """Iterate over terms.

        Deprecated.
        """
        yield from self.read()

    def read(self):
        """Alias for __iter__ (for now)."""
        with open(self._filepath, 'r') as f:
            record = {}

            for line in f:
                if "=" not in line:
                    continue

                key, value = [p.strip() for p in line.split("=", maxsplit=1)]
                self.update_record(record, key, value)

                # Assumes always ends with UI
                if key == "UI":
                    if self._filter(record):
                        yield record
                    record = {}

    def update_record(cls, record, key, value):
        """Update the value of a key in the MeSH record.

        We deal with special cases of interest here.
        """
        if key == "AQ":
            record[key] = [q.strip() for q in value.split()]
        else:
            record[key] = value


class MeSHNewReader:
    """Parser for MeSH new file."""

    def __init__(self, filepath):
        """Constructor."""
        self.filepath = filepath

    def read(self):
        """Read iteratively 'meshnew' file.

        This is a very simple parser that takes advantage of the
        structure of 'meshnew' files.

        Design:
        - lazy load file to reduce memory footprint during read
        """
        with open(self.filepath, 'r') as f:
            entry = {}
            for line in f:
                line = line.strip()
                # partition always yields 3-tuple, values may be '' though
                lv, dash, rv = line.partition("-")
                if dash == "-":
                    entry.update({lv.strip(): rv.strip()})
                else:
                    if entry:
                        yield entry
                        entry = {}
            if entry:
                yield entry


class MeSHReplaceReader:
    """Parser for MeSH replace file."""

    def __init__(self, filepath):
        """Constructor."""
        self.filepath = filepath

    def read(self):
        """Read 'replace' file.

        This is a very simple parser that takes advantage of the
        structure of 'replace' files.

        Design:
        - lazy load file to reduce memory footprint during read
        Output:
            yield
            {
                "MH OLD": "...",
                "MH NEW": "...",
                "delete": "...",
                "status": "...",
            }
        """
        with open(self.filepath, 'r') as f:
            entry = {}
            for line in f:
                line = line.strip()
                valid_start = (
                    line.startswith("MH OLD =") or
                    line.startswith("MH NEW =")
                )
                if valid_start:
                    lv, eq, rv = line.partition("=")
                    lv = lv.strip()
                    rv = rv.strip()

                    if line.startswith("MH OLD ="):
                        m = re.search(r"(#?) \[(\w\*?)?\]$", rv)
                        if not m:
                            raise Exception(f"Malformed line: {line}")
                        piece = {
                            lv: rv[:m.start()].strip(),
                            "delete": m.group(1) or "",
                            "status": m.group(2) or "",
                        }
                    else:
                        piece = {lv: rv}

                    entry.update(piece)

                else:
                    if entry:
                        yield entry
                        entry = {}

            if entry:
                yield entry



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test MeSH-related conversion functionality."""

from galter_subjects_utils.contrib.mesh.converter import MeSHRDMConverter, \
    MeSHSubjectDeltasConverter, Subject


def test_converter():
    topics = [{
        'MH': 'Seed Bank',
        'DC': '1',
        'AQ': ['CL', 'EC'],
        'UI': 'D000068098'
    }]
    qualifiers_mapping = {
        "CL": {
            "QA": "CL",
            "SH": "classification",
            "UI": "Q000145"
        },
        "EC": {
            "QA": "EC",
            "SH": "economics",
            "UI": "Q000191"
        }
    }

    converter = MeSHRDMConverter(topics, qualifiers_mapping)
    objects = [o for o in converter]

    expected = [
        {
            "id": 'https://id.nlm.nih.gov/mesh/D000068098',
            "scheme": "MeSH",
            "subject": "Seed Bank"
        },
        {
            "id": 'https://id.nlm.nih.gov/mesh/D000068098Q000145',
            "scheme": "MeSH",
            "subject": "Seed Bank/classification"
        },
        {
            "id": 'https://id.nlm.nih.gov/mesh/D000068098Q000191',
            "scheme": "MeSH",
            "subject": "Seed Bank/economics"
        },
    ]
    assert expected == objects


# MeSHSubjectDeltasConverter tests
# ===

# Helpers


def subject_unqualified():
    """Subject without qualifier."""
    return Subject(id="D000068098", label="Seed Bank")


def subject_qualified():
    """Subject with qualifier."""
    return Subject(id="D044127Q000941", label="Epigenesis, Genetic/ethics")


def subject_unqualified_present_only_in_dst():
    """Subject: unqualified + present only in dst."""
    return Subject(id="D000092002", label="Document Analysis")


def subject_qualified_present_only_in_dst():
    """Subject: qualified + present only in dst.

    This is also a topic that is not identified as being new in meshnew file
    or modified in replace file. To be discovered, the d<year> file would have
    to be read.
    """
    return Subject(id="D005654Q000187", label="Fundus Oculi/drug effects")


def subjects_renamed():
    """Subject and its renamed version."""
    return (
        Subject(
            id="D044467",
            label="American Native Continental Ancestry Group"
        ),
        Subject(id="D044467", label="American Indian or Alaska Native"),
    )


def subjects_qualified_renamed():
    """Qualified subject and its renamed version."""
    return (
        Subject(
            id="D044467Q000145",
            label="American Native Continental Ancestry Group/classification"
        ),
        Subject(
            id="D044467Q000145",
            label="American Indian or Alaska Native/classification"
        ),
    )


def subjects_qualified_replaced_already_present():
    """Subject and its replaced version.

    This function doesn't enforce an already present replaced version, but
    rather expects it to be used this way.
    The replaced version should be passed as part of the original subjects.
    """
    return (
        Subject(
            id="D018290Q000145",
            label="Cervical Intraepithelial Neoplasia/classification"
        ),
        Subject(
            id="D002578Q000145",
            label="Uterine Cervical Dysplasia/classification"
        )
    )


def subjects_replaced_newly_added():
    """Return a subject and its replaced version.

    This function doesn't enforce a newly added version, but
    rather expects it to be used this way. The replacement subject should not
    be passed as part of the original subjects.

    Subjects are made up because didn't find actual cases in last 3 years.
    """
    return (
        Subject(id="foo-before", label="Foo before"),
        Subject(id="foo-after", label="Foo after")
    )


def subjects_replaced_renamed():
    """Return a subject and its replaced version for the relabelled scenario.

    This is to be used for the case where a subject is replaced by
    a pre-existing but relabelled subject.
    """
    return (
        Subject(
            id="D018290Q000145",
            label="American Indians or Alaska Natives/classification"
        ),
        Subject(
            id="D044467Q000145",
            label="American Indian or Alaska Native/classification"
        )
    )


# Tests


def test_converter_add():
    """Test only addition scenarios."""
    src = [
        subject_unqualified()
    ]
    dst = [
        subject_unqualified(),
        subject_unqualified_present_only_in_dst(),
        subject_qualified_present_only_in_dst()
    ]
    replacement = {}

    converter = MeSHSubjectDeltasConverter(src, dst, replacement)

    add_ops = converter.convert()

    expected = [
        # delta op that could have been generated from just looking at
        # meshnew file
        {
            "type": "add",
            "id": "https://id.nlm.nih.gov/mesh/D000092002",
            "scheme": "MeSH",
            "subject": "Document Analysis"
        },
        # delta op that could not have been generated from just looking at
        # meshnew file
        {
            "type": "add",
            "id": "https://id.nlm.nih.gov/mesh/D005654Q000187",
            "scheme": "MeSH",
            "subject": "Fundus Oculi/drug effects"
        },
    ]
    assert expected == add_ops


def test_converter_rename():
    """Test rename scenarios only."""
    orig_subject, renamed_subject = subjects_renamed()
    src = [
        orig_subject,
        subject_unqualified()
    ]
    dst = [
        subject_unqualified(),
        renamed_subject,
    ]
    replacement = {}
    converter = MeSHSubjectDeltasConverter(src, dst, replacement)

    rename_ops = converter.convert()

    expected = [
        {
            "id": "https://id.nlm.nih.gov/mesh/D044467",
            "type": "rename",
            "scheme": "MeSH",
            "subject": "American Native Continental Ancestry Group",
            "new_subject": "American Indian or Alaska Native"
        },
    ]
    assert expected == rename_ops


def test_converter_replace():
    """Test replace scenarios only (sort-of)."""
    subject_replace_present_src, subject_replace_present_dst = (
        subjects_qualified_replaced_already_present()
    )
    subject_replace_add_src, subject_replace_add_dst = (
        subjects_replaced_newly_added()
    )
    subject_rename_src, subject_rename_dst = subjects_qualified_renamed()
    subject_replace_rename_src, subject_replace_rename_dst = (
        subjects_replaced_renamed()
    )
    src = [
        subject_replace_present_src,
        subject_unqualified(),
        subject_replace_present_dst,
        subject_replace_add_src,
        subject_replace_rename_src,
        subject_rename_src,
    ]
    dst = [
        subject_unqualified(),
        subject_replace_present_dst,
        subject_replace_add_dst,
        subject_replace_rename_dst,
        subject_rename_dst,
    ]
    replacements = {
        "Cervical Intraepithelial Neoplasia": "Uterine Cervical Dysplasia",
        "Foo before": "Foo after",
        "American Indians or Alaska Natives": "American Indian or Alaska Native",  # noqa
    }
    converter = MeSHSubjectDeltasConverter(src, dst, replacements)

    ops = converter.convert()

    expected = [
        # contains add and rename operations to account for replacement
        # scenarios
        {
            "id": "https://id.nlm.nih.gov/mesh/foo-after",
            "type": "add",
            "scheme": "MeSH",
            "subject": "Foo after"
        },
        {
            "id": "https://id.nlm.nih.gov/mesh/D044467Q000145",
            "type": "rename",
            "scheme": "MeSH",
            "subject": "American Native Continental Ancestry Group/classification",  # noqa
            "new_subject": "American Indian or Alaska Native/classification"
        },
        {
            "id": "https://id.nlm.nih.gov/mesh/D018290Q000145",
            "type": "replace",
            "scheme": "MeSH",
            "subject": "Cervical Intraepithelial Neoplasia/classification",
            "new_id": "https://id.nlm.nih.gov/mesh/D002578Q000145"
        },
        {
            "id": "https://id.nlm.nih.gov/mesh/foo-before",
            "type": "replace",
            "scheme": "MeSH",
            "subject": "Foo before",
            "new_id": "https://id.nlm.nih.gov/mesh/foo-after"
        },
        {
            "id": "https://id.nlm.nih.gov/mesh/D018290Q000145",
            "type": "replace",
            "scheme": "MeSH",
            "subject": "American Indians or Alaska Natives/classification",
            "new_id": "https://id.nlm.nih.gov/mesh/D044467Q000145"
        },
    ]
    assert expected == ops


def test_converter_remove():
    """Test remove scenarios only."""
    subject = subject_qualified()
    src = [subject]
    dst = []
    replacements = {}
    converter = MeSHSubjectDeltasConverter(src, dst, replacements)

    ops = converter.convert()

    expected = [
        {
            "id": "https://id.nlm.nih.gov/mesh/" + subject.id,
            "type": "remove",
            "scheme": "MeSH",
            "subject": "Epigenesis, Genetic/ethics",
        },
    ]
    assert expected == ops



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023-2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Command line tool."""

from datetime import date
from pathlib import Path

import click
from flask.cli import with_appcontext

from .contrib.mesh.cli import mesh
from .converter import LCSHRDMConverter
from .downloader import LCSHDownloader
from .keeptrace import KeepTrace
from .reader import read_csv, read_jsonl
from .updater import SubjectDeltaUpdater
from .writer import SubjectDeltaLogger, write_jsonl


@click.group()
def main():
    """A subjects CLI utility (mostly for InvenioRDM)."""


main.add_command(mesh)


defaults = {
    "year": date.today().year,
    "filter": "topic-qualifier",
    "downloads-dir": Path.cwd(),
    "output-file": Path.cwd(),
}


def to_lcsh_downloader_kwargs(parameters):
    """To LCSHDownloader kwargs."""
    result = {
        "cache": not parameters["no_cache"],
        "directory": Path.cwd() / parameters["downloads_dir"]
    }
    return result


def to_lcsh_converter_kwargs(downloader):
    """Provide LCSH converter args from `downloader`."""
    result = {
        "topics": read_jsonl(downloader.extracted_filepath)
    }
    return result


@main.command()
@click.option(
    "--downloads-dir", "-d",
    type=click.Path(path_type=Path),
    default=defaults["downloads-dir"])
@click.option(
    "--output-file", "-o",
    type=click.Path(path_type=Path),
    default=defaults["output-file"] / "subjects_lcsh.jsonl",
)
@click.option("--no-cache", default=False)
def lcsh(**parameters):
    """LCSH related commands."""
    # Download
    downloader_kwargs = to_lcsh_downloader_kwargs(parameters)
    downloader = LCSHDownloader(**downloader_kwargs)
    downloader.download()

    # Convert
    converter_kwargs = to_lcsh_converter_kwargs(downloader)
    converter = LCSHRDMConverter(**converter_kwargs)

    # Write
    filepath = write_jsonl(converter, parameters["output_file"])

    print(f"LCSH terms written here {filepath}")


keep_trace_field_help = "Dotted field path to where trace should be kept."
keep_trace_tmpl_help = "Template with expandable '{subject}' to be saved."


@main.command("update")
@click.argument(
    "deltas-file",
    type=click.Path(path_type=Path, exists=True, dir_okay=False),
)
@click.option(
    "--output-file", "-o",
    type=click.Path(path_type=Path),
    default=defaults["output-file"] / "updated_records.csv",
)
@click.option("--keep-trace-field", "-f", help=keep_trace_field_help)
@click.option("--keep-trace-template", "-t", help=keep_trace_tmpl_help)
@with_appcontext
def update_subjects(**parameters):
    """Update subjects in running instance according to deltas file."""
    print(f"Updating subjects...")
    deltas = [d for d in read_csv(parameters["deltas_file"])]
    log_filepath = parameters["output_file"]
    logger = SubjectDeltaLogger(filepath=log_filepath)
    keep_trace = KeepTrace(
        field=parameters.get("keep_trace_field") or None,
        template=parameters.get("keep_trace_template") or None
    )
    updater = SubjectDeltaUpdater(deltas, logger, keep_trace)
    updater.update()
    print(f"Log of updated records written here {log_filepath}")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""KeepTrace."""

from dataclasses import dataclass


@dataclass
class KeepTrace:
    """Keeps trace of subject at field in record using template."""

    field: str  # dotted path to field
    template: str
    yes = "Y"
    no = "N"
    keep_trace_key = "keep_trace"

    @classmethod
    def mark(cls, ops_data, yes_logic):
        """Mark ops_data in-place to keep trace or not depending on yes_logic.

        :param ops_data: operation data
        :type ops_data: List[dict] (list of subject op data)
        :param yes_logic: function to determine if op_data should be marked
        :type yes_logic: function(op_data) -> Bool
        """
        for op_data in ops_data:
            op_data[cls.keep_trace_key] = cls.yes if yes_logic(op_data) else cls.no  # noqa

    def should_trace(self, op_data):
        """Determine if op_data should trace."""
        return op_data.get(self.keep_trace_key, "").lower() == self.yes.lower()

    def trace(self, record, subject):
        """Save expanded `self.template` at `self.field` in record."""
        if not self.field or not self.template or not subject:
            return

        final_dict = self.find_final_dict(record)
        self.assign_template(final_dict, subject)

    def find_final_dict(self, record):
        """Find or create final dict by following `field`."""
        obj = record
        keys = self.field.split(".")

        for key in keys[:-1]:
            got = obj.get(key)
            if isinstance(got, dict):
                obj = got
            elif got is None:
                new_dict = {}
                obj[key] = new_dict
                obj = new_dict
            elif isinstance(got, list):
                new_dict = {}
                got.append(new_dict)
                obj = new_dict
            else:
                break

        assert isinstance(obj, dict), f"KeepTrace.field '{self.field}' is invalid."  # noqa
        return obj

    def assign_template(self, dict_, subject):
        """Assign expanded template."""
        final_key = self.field.split(".")[-1]
        dict_[final_key] = self.template.format(subject=subject)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2022 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""MeSH subjects_mesh.jsonl writer."""

import csv
import json
from datetime import datetime
from io import StringIO
from pathlib import Path


def write_jsonl(
    entries,
    filepath=Path(__file__).parent / "vocabularies/subjects_mesh.jsonl"
):
    """Write the MeSH jsonl file.

    Return filepath to written file.
    """
    with open(filepath, "w") as f:
        for entry in entries:
            json.dump(entry, f)
            f.write("\n")

    return filepath


def write_csv(entries, filepath, writer_kwargs=None):
    """Write to CSV."""
    writer_kwargs = writer_kwargs or {}
    with open(filepath, "w") as f:
        writer = csv.DictWriter(f, **writer_kwargs)
        writer.writeheader()
        writer.writerows(entries)

    return filepath


class SubjectDeltaLogger:
    """Convenience logger for delta operations applied to records."""

    def __init__(self, filepath=None):
        """Constructor."""
        if not filepath:
            # In memory
            self.f = StringIO()
        else:
            self.f = open(filepath, "w+", newline='')

        self.header = ["pid", "time", "error", "deltas"]
        self.writer = csv.DictWriter(
            self.f, fieldnames=self.header
        )
        self.writer.writeheader()

        self.clear()

    def log(self, pid, delta=None, error=None):
        """Buffered log.

        Idea is to build up a log entry for pid with multiple deltas.
        The actual logging is only written out when flush() is called.
        An impromptu change in pid will clear out any accumulated data.
        `log(error=...)` is typically called on its own to fill the error
        portion.
        """
        if self.pid != pid:
            self.clear()

        self.pid = pid

        if delta:
            if delta["type"] == "replace":
                delta_msg = f"{delta['id']} -> {delta['new_id']}"
            elif delta["type"] == "remove":
                delta_msg = f"{delta['id']} -> X"
            elif delta["type"] == "rename":
                delta_msg = f"{delta['id']} {delta['subject']} -> {delta['new_subject']}"  # noqa
            else:
                return

            self.deltas.append(delta_msg)

        if error:
            self.error = error

    def flush(self):
        """Flush out the buffered entry."""
        if not self.pid:
            # do nothing
            return

        self.writer.writerow(
            {
                "pid": self.pid,
                "time": datetime.now(),
                "error": self.error,
                "deltas": " + ".join(self.deltas)
            }
        )

        self.clear()

    def clear(self):
        """Clear out the buffered data."""
        self.pid = None
        self.error = None
        self.deltas = []

    def close(self):
        """Close file."""
        self.f.close()

    def read(self):
        """Read file."""
        offset = self.f.tell()
        self.f.seek(0)
        result = [e for e in csv.DictReader(self.f)]
        self.f.seek(offset)
        return result



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023-2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""MeSH Command line tool."""

from datetime import date
from functools import wraps
from pathlib import Path

import click
from flask.cli import with_appcontext

from galter_subjects_utils.keeptrace import KeepTrace
from galter_subjects_utils.reader import get_rdm_subjects, mapping_by
from galter_subjects_utils.writer import write_csv, write_jsonl

from .adapter import converted_to_subjects, generate_replacements
from .converter import MeSHRDMConverter, MeSHSubjectDeltasConverter
from .downloader import MeSHDownloader
from .reader import MeSHReader, MeSHReplaceReader, topic_filter

defaults = {
    "year": date.today().year,
    "filter": "topic-qualifier",
    "downloads-dir": Path.cwd(),
    "output-file": Path.cwd(),
}


def to_mesh_downloader_kwargs(parameters):
    """To MeSHDownloader kwargs."""
    result = {
        "year": parameters["year"],
        "cache": not parameters["no_cache"],
        "directory": Path.cwd() / parameters["downloads_dir"]
    }

    return result


def to_mesh_converter_kwargs(parameters, downloader):
    """Convert."""
    result = {}

    if parameters["filter"] in ["topic", "topic-qualifier"]:
        result["topics"] = (
            MeSHReader(
                downloader.terms_filepath,
                filter=topic_filter,
            )
            .read()
        )
    if parameters["filter"] in ["topic-qualifier"]:
        result["qualifiers_mapping"] = (
            mapping_by(
                MeSHReader(downloader.qualifiers_filepath).read(),
                by="QA"
            )
        )

    return result


def mesh_download_options(f):
    """Encapsulate common MeSH download options."""

    @click.option("--year", "-y", default=defaults["year"])
    @click.option(
        "--filter", "-f",
        type=click.Choice(["topic", "topic-qualifier"]),
        default=defaults["filter"],
    )
    @click.option(
        "--downloads-dir", "-d",
        type=click.Path(path_type=Path),
        default=defaults["downloads-dir"]
    )
    @click.option(
        "--no-cache",
        default=False,
        help="Re-download even if already downloaded.",
        is_flag=True
    )
    @wraps(f)
    def _wrapped(*args, **kwargs):
        """Wrap f with common download options."""
        return f(*args, **kwargs)

    return _wrapped


@click.group()
def mesh():
    """MeSH related commands."""


@mesh.command("download")
@mesh_download_options
def mesh_download(**parameters):
    """Download MeSH files."""
    downloader_kwargs = to_mesh_downloader_kwargs(parameters)
    downloader = MeSHDownloader(**downloader_kwargs)
    downloader.download()

    print(f"Raw MeSH files written in {parameters['downloads_dir']}/")


@mesh.command("file")
@mesh_download_options
@click.option(
    "--output-file", "-o",
    type=click.Path(path_type=Path),
    default=defaults["output-file"] / "subjects_mesh.jsonl",
)
def mesh_file(**parameters):
    """Generate new MeSH subjects file."""
    # Download
    downloader_kwargs = to_mesh_downloader_kwargs(parameters)
    downloader = MeSHDownloader(**downloader_kwargs)
    downloader.download()

    # Convert
    converter_kwargs = to_mesh_converter_kwargs(parameters, downloader)
    converter = MeSHRDMConverter(**converter_kwargs)

    # Write
    header = ["id", "scheme", "subject"]
    filepath = write_csv(
        converter.convert(),
        parameters["output_file"],
        writer_kwargs={
            "fieldnames": header
        }
    )

    print(f"MeSH terms written here {filepath}")


@mesh.command("deltas")
@click.option(
    "--downloads-dir", "-d",
    type=click.Path(path_type=Path),
    default=defaults["downloads-dir"])
@click.option("--year", "-y", type=int, default=defaults["year"])
@click.option(
    "--filter", "-f",
    type=click.Choice(["topic", "topic-qualifier"]),
    default=defaults["filter"],
)
@click.option(
    "--output-file", "-o",
    type=click.Path(path_type=Path),
    default=defaults["output-file"] / "deltas.csv",
)
@with_appcontext
def mesh_deltas(**parameters):
    """Write MeSH subject delta operations to file."""
    print("Generating deltas...")
    downloads_dir = parameters["downloads_dir"].expanduser()
    year = parameters["year"]
    filter_ = parameters["filter"]

    # Source subjects
    subject_rdm_preexisting = get_rdm_subjects(scheme="MeSH")
    src = converted_to_subjects(subject_rdm_preexisting)

    # Destination subjects
    subjects_fp = downloads_dir / f"d{year}.bin"
    topics = MeSHReader(subjects_fp, filter=topic_filter).read()

    if filter_ == "topic-qualifier":
        qualifiers_fp = downloads_dir / f"q{year}.bin"
        qualifiers_mapping = mapping_by(
            MeSHReader(qualifiers_fp).read(),
            by="QA"
        )
    else:
        qualifiers_mapping = {}

    converted = MeSHRDMConverter(topics, qualifiers_mapping).convert()
    dst = converted_to_subjects(converted)

    # Replacements
    replace_fp = downloads_dir / f"replace{year}.txt"
    replacements = generate_replacements(
        MeSHReplaceReader(replace_fp).read()
    )

    ops = (
        MeSHSubjectDeltasConverter(
            src_subjects=src,
            dst_subjects=dst,
            replacements=replacements
        )
        .convert()
    )
    # in-place
    KeepTrace.mark(
        ops,
        # only can keep trace of *those* anyway
        yes_logic=lambda op: op["type"] in ["rename", "replace", "remove"]
    )

    deltas_fp = parameters["output_file"]
    header = [
        "id",
        "type",
        "scheme",
        "subject",
        "new_id",
        "new_subject",
        "keep_trace"
    ]
    write_csv(
        ops,
        deltas_fp,
        writer_kwargs={
            "fieldnames": header
        }
    )

    print(f"MeSH deltas written here {deltas_fp}")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test MeSH-related reading functionality."""

from pathlib import Path

from galter_subjects_utils.contrib.mesh.reader import MeSHNewReader, \
    MeSHReader, MeSHReplaceReader, topic_filter

# Utilities


def assert_includes(dicts, dict_cores):
    """Checks that each dict in dicts has the corresponding dict_core."""
    for d, dc in zip(dicts, dict_cores):
        for key, value in dc.items():
            assert value == d[key]

# Tests


def test_reader_descriptors_filter():
    filepath = Path(__file__).parent / "data" / "fake_d2022.bin"

    reader = MeSHReader(filepath, filter=topic_filter)
    topics = [t for t in reader]

    expected_cores = [
        {
            'MH': 'Seed Bank',
            'DC': '1',
            'AQ': ['CL', 'EC'],
            'UI': 'D000068098'
        },
        {
            'MH': 'Abbreviations as Topic',
            'DC': '1',
            'UI': 'D000004'
        },
        {
            'MH': 'Abdomen',
            'DC': '1',
            'AQ': ['AB', 'AH'],
            'UI': 'D000005'
        },
        {
            'MH': 'American Indians or Alaska Natives',
            'DC': '1',
            'AQ': ['CL', 'ED', 'EH', 'GE', 'HI', 'LJ', 'PX', 'SN'],
            'UI': 'D000086562'
        }
    ]
    assert_includes(topics, expected_cores)


def test_reader_qualifiers():
    filepath = Path(__file__).parent / "data" / "fake_q2023.bin"

    reader = MeSHReader(filepath)
    qualifiers = [q for q in reader]

    expected_cores = [
        {
            "QA": "AB",
            "SH": "abnormalities",
            "UI": "Q000002"
        },
        {
            "QA": "AH",
            "SH": "anatomy & histology",
            "UI": "Q000033"
        },
        {
            "QA": "CL",
            "SH": "classification",
            "UI": "Q000145"
        },
        {
            "QA": "EC",
            "SH": "economics",
            "UI": "Q000191"
        },
    ]
    assert_includes(qualifiers, expected_cores)


def test_new_reader():
    filepath = Path(__file__).parent / "data" / "fake_meshnew2023.txt"

    reader = MeSHNewReader(filepath)
    entries = [e for e in reader.read()]

    expected = [
        {
            "MH": "Document Analysis",
            "UI": "D000092002",
            "MN": "H1.770.644.241.850.375",
            "MS": "A form of qualitative research that uses a systematic procedure to analyze documentary evidence and answer specific research questions.",  # noqa
            "HN": "2023",
        },
        {
            "MH": "Family Structure",
            "UI": "D000092822",
            "MN": "F1.829.263.315.250",
            "MN": "I1.240.361.330",
            "MN": "I1.880.761.125",
            "MN": "I1.880.853.150.423.250",
            "MN": "N1.224.361.330",
            "MN": "N1.824.308.125",
            "MN": "N6.850.505.400.400.580",
            "MS": "Structural nature of relationships among members of a household typically in reference to a MINOR residing in the same home. More broadly any organizational framework that determines family membership, and the functions and hierarchical position of family members (https://eric.ed.gov/?qt=Family+Structure&ti=Family+Structure).",  # noqa
            "HN": "2023; for STEPFAMILY and FAMILY, RECONSTITUTED use FAMILY 1996-2022; for MATRIARCHY and PATRIARCHY use FAMILY CHARACTERISTICS 1995-2022",  # noqa
            "BX": "Family, Reconstituted",
            "BX": "Reconstituted Family",
            "BX": "Step-parent Family",
            "BX": "Stepfamily",
            "BX": "Stepparent Family",
        }
    ]
    assert expected == entries


def test_replace_reader():
    filepath = Path(__file__).parent / "data" / "fake_replace2023.txt"

    reader = MeSHReplaceReader(filepath)
    entries = [e for e in reader.read()]

    expected = [
        {
            "MH OLD": "Far East",
            "MH NEW": "Asia, Eastern",
            "delete": "",
            "status": "P",
        },
        {
            "MH OLD": "American Indians or Alaska Natives",
            "MH NEW": "American Indian or Alaska Native",
            "delete": "#",
            "status": "P",
        },
        {
            "MH OLD": "Asians",
            "MH NEW": "Asian People",
            "delete": "",
            "status": "N*",
        },
        {
            "MH OLD": "Whites",
            "MH NEW": "White People",
            "delete": "",
            "status": "P*",
        },
        {
            "MH OLD": "RNA, Guide",
            "MH NEW": "RNA, Guide, Kinetoplastida",
            "delete": "",
            "status": "",
        }
    ]
    assert expected == entries



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2023 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Generic download functionality."""

import gzip
import shutil

import requests


def download_file(url, dest):
    """Download a file."""
    with requests.get(url, stream=True) as req:
        with open(dest, 'wb') as f:
            shutil.copyfileobj(req.raw, f)

    return dest


def extract_file(in_filepath, out_filepath):
    """Extract gzipped file."""
    with gzip.open(in_filepath, 'rb') as f_in:
        with open(out_filepath, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)

    return out_filepath


class LCSHDownloader:
    """Download LCSH file."""

    def __init__(self, directory, cache=False):
        """Constructor."""
        self.base_url = "https://id.loc.gov/download/authorities/subjects.skosrdf.jsonld.gz"  # noqa
        self.directory = directory
        self.filepath = self.directory / self.base_url.rsplit("/")[-1]
        self.cache = cache

    def download(self):
        """Download LCSH files of interest."""
        if not self.cache or not self.filepath.exists():
            download_file(self.base_url, self.filepath)
            # sneak extraction in caching: bit of a cheat, but ok for now
            extract_file(self.filepath, self.filepath.with_suffix(""))

        self.downloaded_filepath = self.filepath  # just a marker + future
        self.extracted_filepath = self.filepath.with_suffix("")
        return self.extracted_filepath



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Generic reader functionality."""

import csv
import json

from invenio_db import db
from invenio_vocabularies.contrib.subjects.models import SubjectsMetadata
from sqlalchemy import bindparam, select, text


def read_jsonl(filepath):
    """KISS jsonl file reader."""
    with open(filepath) as f:
        for line in f:
            yield json.loads(line)


def read_csv(filepath, reader_kwargs=None):
    """KISS csv reader."""
    reader_kwargs = reader_kwargs or {}
    with open(filepath) as f:
        reader = csv.DictReader(f, **reader_kwargs)
        yield from reader


def mapping_by(iterable, by, keys=None):
    """Return dict out of `iterable` mapped by `by` with only `keys` chosen.

    If multiple entries in iterable have same `by` value, the last one will
    take precedence.

    :param iterable: iterable of dict
    :param by: key used to group
    :param keys: keys of each dict in iterable that should be kept
                 (with their values)
    :return: dict
    """
    return {
        d.get(by): {k: d.get(k) for k in keys} if keys else d
        for d in iterable
    }


def get_rdm_subjects(scheme):
    """Return all rdm subjects of corresponding scheme."""
    is_scheme = (
        text('json::json->>\'scheme\' = :scheme')
        .bindparams(
            bindparam(
                "scheme",
                value=scheme,
            )
        )
    )

    stmt = (
        select(SubjectsMetadata.json)
        .where(is_scheme)
    )

    return db.session.scalars(stmt)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test KeepTrace."""

from galter_subjects_utils.keeptrace import KeepTrace


def test_mark():
    delta_ops = [
        {
            "type": "remove",
            "id": "http://example.org/baz/0",
            "scheme": "baz",
            "subject": "0",
        },
        {
            "type": "replace",
            "id": "http://example.org/baz/1",
            "scheme": "baz",
            "subject": "1",
            "new_id": "http://example.org/baz/2",
        },
        {
            "type": "rename",
            "id": "http://example.org/baz/2",
            "scheme": "baz",
            "subject": "2",
            "new_subject": "Baz-Two",
        }
    ]

    KeepTrace.mark(
        delta_ops,
        yes_logic=lambda o: o["type"] in ["remove", "rename"]
    )

    assert "Y" == delta_ops[0][KeepTrace.keep_trace_key]
    assert "N" == delta_ops[1][KeepTrace.keep_trace_key]
    assert "Y" == delta_ops[2][KeepTrace.keep_trace_key]


def test_should_trace():
    op_no_key = {
        "type": "rename",
        "id": "http://example.org/baz/2",
        "scheme": "baz",
        "subject": "2",
        "new_subject": "Baz-Two",
    }
    keep_trace = KeepTrace(field="any", template="any")

    assert keep_trace.should_trace(op_no_key) is False

    op_random_value = {**op_no_key, KeepTrace.keep_trace_key: "random"}
    assert keep_trace.should_trace(op_random_value) is False

    op_yes = {**op_no_key, KeepTrace.keep_trace_key: "Y"}
    assert keep_trace.should_trace(op_yes)

    op_lower = {**op_no_key, KeepTrace.keep_trace_key: "y"}
    assert keep_trace.should_trace(op_lower)

    op_no = {**op_no_key, KeepTrace.keep_trace_key: "n"}
    assert keep_trace.should_trace(op_no) is False



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test updater."""

import copy

import pytest
from invenio_access.permissions import system_identity
from invenio_pidstore.errors import PIDDoesNotExistError
from invenio_rdm_records.records import RDMDraft, RDMRecord
from invenio_records_resources.proxies import current_service_registry
from invenio_vocabularies.contrib.subjects.api import Subject

from galter_subjects_utils.keeptrace import KeepTrace
from galter_subjects_utils.updater import SubjectDeltaUpdater
from galter_subjects_utils.writer import SubjectDeltaLogger


# Fixtures
@pytest.fixture(scope="module")
def subjects_service():
    return current_service_registry.get("subjects")


@pytest.fixture(scope="module")
def subjects(app, subjects_service):
    """A couple of subject records."""
    result = [
        subjects_service.create(
            system_identity,
            {
                "id": "http://example.org/foo/0",
                "scheme": "foo",
                "subject": "0",
            },
        ),
        subjects_service.create(
            system_identity,
            {
                "id": "http://example.org/foo/1",
                "scheme": "foo",
                "subject": "1",
            },
        ),
        subjects_service.create(
            system_identity,
            {
                "id": "http://example.org/foo/2",
                "scheme": "foo",
                "subject": "2",
            },
        ),
        subjects_service.create(
            system_identity,
            {
                "id": "http://example.org/bar/0",
                "scheme": "bar",
                "subject": "0",
            },
        )
    ]

    # Need to make sure records are indexed before subsequent interactions
    Subject.index.refresh()

    return result


@pytest.fixture(scope="module")
def operations_data():
    """Operations in data form."""
    return [
        {
            "type": "add",
            "scheme": "foo",
            "id": "http://example.org/foo/4",
            "subject": "4"
        },
        {
            "type": "rename",
            "scheme": "foo",
            "id": "http://example.org/foo/1",
            "subject": "1",
            "new_subject": "One"
        },
        {
            "type": "remove",
            "scheme": "foo",
            "id": "http://example.org/foo/0",
            "subject": "0"
        },
        {
            "type": "replace",
            "scheme": "foo",
            "id": "http://example.org/foo/2",
            "subject": "2",
            "new_id": "http://example.org/foo/4",
            "new_subject": "4"
        },
    ]


@pytest.fixture(scope="module")
def records_data_w_subjects(
    create_draft_data, create_record_data, minimal_record_input
):
    """Records tagged with subjects."""
    result = []

    record_input = copy.deepcopy(minimal_record_input)
    record_input["metadata"]["subjects"] = [
        {"id": "http://example.org/foo/1"},  # will be renamed
        {"id": "http://example.org/foo/2"},  # will be replaced
        {"id": "http://example.org/bar/0"},  # will be left alone
        {"subject": "a_keyword"},
    ]
    result += [create_record_data(system_identity, record_input)]

    record_input = copy.deepcopy(minimal_record_input)
    record_input["metadata"]["subjects"] = [
        {"id": "http://example.org/foo/0"},  # will be removed
    ]
    result += [create_record_data(system_identity, record_input)]

    # draft
    record_input = copy.deepcopy(minimal_record_input)
    record_input["metadata"]["subjects"] = [
        {"id": "http://example.org/foo/2"},  # will be replaced
    ]
    result += [create_draft_data(system_identity, record_input)]

    return result


@pytest.fixture(scope="module")
def delta_logger():
    """Log record deltas."""
    return SubjectDeltaLogger()


@pytest.fixture(scope="module")
def update_result(
    running_app, subjects, records_data_w_subjects, operations_data,
    delta_logger
):
    """Update results.

    This is actually the code under test. Because it does a lot, we create
    multiple tests that all test different aspects of its run.
    """
    noop_keep_trace = KeepTrace(None, None)
    executor = SubjectDeltaUpdater(
        operations_data, delta_logger, noop_keep_trace)
    result = executor.update()

    # Really just needed for tests: OS indices are refreshed on an interval
    # longer than the queries we make in tests, so we have to force a refresh.
    RDMDraft.index.refresh()
    RDMRecord.index.refresh()
    Subject.index.refresh()

    return result


# Test utilities
def any_contains(dicts, dict_):
    """Return bool if any dict in `dicts` contains `dict_`."""
    return any(dict_.items() <= d.items() for d in dicts)


def get_subjects_of_record_from_db(pid, draft=False):
    """Return DB subjects of record with `pid`."""
    records_service = current_service_registry.get("records")
    record_result = (
        records_service.read_draft(system_identity, pid)
        if draft else
        records_service.read(system_identity, pid)
    )
    record_dict = record_result.to_dict()
    return record_dict["metadata"]["subjects"]


def get_records_from_de(draft=False):
    """Return DE hits."""
    records_service = current_service_registry.get("records")
    record_results = (
        records_service.search_drafts(system_identity)
        if draft else
        records_service.search(system_identity)
    )
    records_dict = record_results.to_dict()
    return records_dict["hits"]["hits"]


def get_subjects_of_record_from_de(records, pid):
    """Return subjects of record with `pid` in `records` .

    records is the output of `get_records_from_de`.
    """
    record_dict = next((r for r in records if r.get("id") == pid), {})
    return record_dict.get("metadata").get("subjects", [])


# Tests
def test_update_on_subjects(update_result, subjects_service):
    # update_result is under test

    # at DB
    # -----
    # add
    assert subjects_service.read(system_identity, "http://example.org/foo/4")

    # rename
    subject_result = subjects_service.read(
        system_identity, "http://example.org/foo/1"
    )
    subject_dict = subject_result.to_dict()
    assert "One" == subject_dict["subject"]
    # what happens with other changes?

    # remove
    with pytest.raises(PIDDoesNotExistError):
        subjects_service.read(system_identity, "http://example.org/foo/0")

    # replace
    with pytest.raises(PIDDoesNotExistError):
        subjects_service.read(system_identity, "http://example.org/foo/2")

    # at index
    # --------
    # fields = None should give all fields
    subject_results = subjects_service.read_all(system_identity, fields=None)
    subjects_dict = subject_results.to_dict()
    hits = subjects_dict["hits"]["hits"]

    # add
    assert any_contains(hits, {"id": "http://example.org/foo/4"})
    # rename
    assert any_contains(hits, {"subject": "One"})
    # remove
    assert not any_contains(hits, {"id": "http://example.org/foo/0"})
    # replace
    assert not any_contains(hits, {"id": "http://example.org/foo/2"})


def test_update_on_records(update_result, records_data_w_subjects):
    # update_result is under test

    # at DB
    # -----
    # first record
    pid_0 = records_data_w_subjects[0].pid.pid_value
    subjects = get_subjects_of_record_from_db(pid_0)
    assert 4 == len(subjects)
    # replace
    assert not any_contains(subjects, {"id": "http://example.org/foo/2"})
    assert any_contains(subjects, {"id": "http://example.org/foo/4"})
    # rename
    assert any_contains(
        subjects,
        {"id": "http://example.org/foo/1", "subject": "One"},
    )
    # leave alone
    assert any_contains(subjects, {"id": "http://example.org/bar/0"})
    assert any_contains(subjects, {"subject": "a_keyword"})

    # second record
    pid_1 = records_data_w_subjects[1].pid.pid_value
    subjects = get_subjects_of_record_from_db(pid_1)
    # remove
    assert 0 == len(subjects)  # only had the one removed subject

    # at document engine
    # --------
    records = get_records_from_de()
    # first record
    subjects = get_subjects_of_record_from_de(records, pid_0)
    assert 4 == len(subjects)
    # replace
    assert not any_contains(subjects, {"id": "http://example.org/foo/2"})
    assert any_contains(subjects, {"id": "http://example.org/foo/4"})
    # rename
    assert any_contains(
        subjects,
        {"id": "http://example.org/foo/1", "subject": "One"},
    )
    # leave alone
    assert any_contains(subjects, {"id": "http://example.org/bar/0"})
    assert any_contains(subjects, {"subject": "a_keyword"})

    # second record
    subjects = get_subjects_of_record_from_de(records, pid_1)
    # remove
    assert 0 == len(subjects)  # only had the one removed subject


def test_update_on_drafts(update_result, records_data_w_subjects):
    # update_result is under test

    # at DB
    # -----
    pid = records_data_w_subjects[2].pid.pid_value
    subjects = get_subjects_of_record_from_db(pid, draft=True)
    assert not any_contains(subjects, {"id": "http://example.org/foo/2"})
    assert any_contains(subjects, {"id": "http://example.org/foo/4"})

    # at document engine
    # ---
    records = get_records_from_de(draft=True)
    subjects = get_subjects_of_record_from_de(records, pid)
    # Drafts don't keep trace
    assert 1 == len(subjects)
    assert any_contains(subjects, {"id": "http://example.org/foo/4"})


def test_update_logging(update_result, delta_logger, records_data_w_subjects):
    log_entries = delta_logger.read()

    # record 0 logging
    pid = records_data_w_subjects[0].pid.pid_value
    log_entry = next((e for e in log_entries if e.get("pid") == pid), None)
    assert log_entry and log_entry["time"]
    assert "" == log_entry["error"]
    msg = (
        "http://example.org/foo/1 1 -> One + " +
        "http://example.org/foo/2 -> http://example.org/foo/4"
    )
    assert msg == log_entry["deltas"]

    # record 1 logging
    pid = records_data_w_subjects[1].pid.pid_value
    log_entry = next(
        (e for e in log_entries if e.get("pid") == pid),
        None
    )
    assert log_entry["time"]
    assert "" == log_entry["error"]
    msg = "http://example.org/foo/0 -> X"
    assert msg == log_entry["deltas"]

    # draft logging
    pid = records_data_w_subjects[2].pid.pid_value
    log_entry = next(
        (e for e in log_entries if e.get("pid") == pid),
        None
    )
    assert log_entry and log_entry["time"]
    assert "" == log_entry["error"]
    msg = "http://example.org/foo/2 -> http://example.org/foo/4"
    assert msg == log_entry["deltas"]


def test_update_keep_trace(
    create_subject_data, minimal_record_input, create_record_data_fn,
):
    # Test that keep_trace template is placed in keep_trace field
    # if subject delta op is marked as keep_trace
    subjects_data = [
        create_subject_data(
            system_identity,
            {
                "id": f"http://example.org/baz/{i}",
                "scheme": "baz",
                "subject": f"{i}",
            },
        )
        for i in range(3)
    ]
    record_input = copy.deepcopy(minimal_record_input)
    record_input["metadata"]["subjects"] = [
        {
            "id": "http://example.org/baz/0",
        },
        {
            "id": "http://example.org/baz/1",
        }
    ]
    record_0_data = create_record_data_fn(system_identity, record_input)
    record_input = copy.deepcopy(minimal_record_input)
    record_input["metadata"]["subjects"] = [
        {
            "id": "http://example.org/baz/2"
        }
    ]
    record_1_data = create_record_data_fn(system_identity, record_input)
    delta_ops = [
        {
            "type": "remove",
            "id": "http://example.org/baz/0",
            "scheme": "baz",
            "subject": "0",
            "keep_trace": "Y",
        },
        {
            "type": "replace",
            "id": "http://example.org/baz/1",
            "scheme": "baz",
            "subject": "1",
            "new_id": "http://example.org/baz/2",
            "keep_trace": "N",
        },
        {
            "type": "rename",
            "id": "http://example.org/baz/2",
            "scheme": "baz",
            "subject": "2",
            "new_subject": "Baz-Two",
            "keep_trace": "Y",
        }
    ]
    delta_logger = SubjectDeltaLogger()
    keep_trace = KeepTrace(
        field="metadata.subjects.subject",
        template="anything {subject} anything"
    )
    # make sure indices are refreshed
    RDMRecord.index.refresh()
    Subject.index.refresh()

    updater = SubjectDeltaUpdater(delta_ops, delta_logger, keep_trace)
    updater.update()

    # at DB
    # -----
    # record 0
    subjects = get_subjects_of_record_from_db(record_0_data.pid.pid_value)
    assert 2 == len(subjects)
    # Contains keep_trace template due to removal
    assert any_contains(subjects, {"subject": "anything 0 anything"})
    # Doesn't contain keep_trace template due to replacement
    assert not any_contains(subjects, {"subject": "anything 1 anything"})
    assert any_contains(subjects, {"id": "http://example.org/baz/2"})

    # record 1
    subjects = get_subjects_of_record_from_db(record_1_data.pid.pid_value)
    assert 2 == len(subjects)
    # Contains keep_trace template due to rename
    assert any_contains(subjects, {"subject": "anything 2 anything"})
    assert any_contains(subjects, {"subject": "Baz-Two"})

    # at document engine
    # ---
    # This is done for completeness sake, just in case some changes are made
    # that would inadvertently skip updating the document engine.
    RDMRecord.index.refresh()
    records = get_records_from_de()

    # record 0
    subjects = get_subjects_of_record_from_de(
        records,
        record_0_data.pid.pid_value
    )
    assert 2 == len(subjects)
    # Contains keep_trace template due to removal
    assert any_contains(subjects, {"subject": "anything 0 anything"})
    # Doesn't contain keep_trace template due to replacement
    assert not any_contains(subjects, {"subject": "anything 1 anything"})
    assert any_contains(subjects, {"id": "http://example.org/baz/2"})

    # record 1
    subjects = get_subjects_of_record_from_de(
        records,
        record_1_data.pid.pid_value
    )
    assert 2 == len(subjects)
    # Contains keep_trace template due to rename
    assert any_contains(subjects, {"subject": "anything 2 anything"})
    assert any_contains(subjects, {"subject": "Baz-Two"})


def test_update_edge_case(
    create_subject_data, minimal_record_input, create_record_data_fn,
):
    # Test edge any other edge case that pops up
    # - subjects of records are deduplicated after modifications

    # Assignments
    subjects_data = [
        create_subject_data(
            system_identity,
            {
                "id": "http://example.org/zim/0",
                "scheme": "zim",
                "subject": "0",
            },
        ),
        create_subject_data(
            system_identity,
            {
                "id": "http://example.org/zim/0QA",
                "scheme": "zim",
                "subject": "0QA",
            },
        ),
    ]

    record_input = copy.deepcopy(minimal_record_input)
    record_input["metadata"]["subjects"] = [
        {"subject": "0QA"},
        {"id": "http://example.org/zim/0"},
        {"id": "http://example.org/zim/0QA"},
    ]
    record_0_data = create_record_data_fn(system_identity, record_input)

    delta_ops = [
        {
            "type": "replace",
            "id": "http://example.org/zim/0QA",
            "scheme": "zim",
            "subject": "0QA",
            "new_id": "http://example.org/zim/0",
            "keep_trace": "Y",
        }
    ]
    delta_logger = SubjectDeltaLogger()
    keep_trace = KeepTrace(
        field="metadata.subjects.subject",
        template="{subject}"
    )
    # make sure indices are refreshed
    RDMRecord.index.refresh()
    Subject.index.refresh()

    # Actions
    updater = SubjectDeltaUpdater(delta_ops, delta_logger, keep_trace)
    updater.update()
    # make sure indices are refreshed
    RDMRecord.index.refresh()
    Subject.index.refresh()

    # Assertions
    # at DB
    # -----
    # record 0
    subjects = get_subjects_of_record_from_db(record_0_data.pid.pid_value)
    assert 2 == len(subjects)
    assert any_contains(subjects, {"subject": "0QA"})
    assert any_contains(subjects, {"id": "http://example.org/zim/0"})

    # at document engine
    # ---
    # This is done for completeness sake, just in case some changes are made
    # that would inadvertently skip updating the document engine.
    RDMRecord.index.refresh()
    records = get_records_from_de()

    # record 0
    subjects = get_subjects_of_record_from_de(
        records,
        record_0_data.pid.pid_value
    )
    assert 2 == len(subjects)
    assert any_contains(subjects, {"subject": "0QA"})
    assert any_contains(subjects, {"id": "http://example.org/zim/0"})



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test generic writer functionality."""

from pathlib import Path

from galter_subjects_utils.reader import read_jsonl
from galter_subjects_utils.writer import SubjectDeltaLogger, write_jsonl


def test_write():
    filepath = Path(__file__).parent / "test_subjects_mesh.jsonl"
    entries = [
        {
            "id": 'D000015',
            "tags": ["mesh"],
            "title": {
                "en": 'Abnormalities, Multiple'
            }
        },
        {
            "id": 'D000068098',
            "tags": ["mesh"],
            "title": {
                "en": 'Seed Bank'
            }
        },
        {
            "id": 'D005368',
            "tags": ["mesh"],
            "title": {
                "en": 'Filariasis'
            }
        }
    ]

    write_jsonl(entries, filepath)

    read_entries = list(read_jsonl(filepath))
    assert entries == read_entries

    filepath.unlink(missing_ok=True)


def test_logging_corner_cases():
    # Log exception messages
    logger = SubjectDeltaLogger()

    delta = {
        "type": "remove",
        "scheme": "foo",
        "id": "A",
    }

    try:
        raise Exception("msg")
    except Exception as e:
        logger.log("abcde-54321", error=str(e))
        logger.flush()

    entries = logger.read()
    assert "msg" == entries[0]["error"]

    # Log multiple deltas
    logger = SubjectDeltaLogger()
    deltas = [
        {
            "type": "remove",
            "scheme": "foo",
            "id": "A",
        },
        {
            "type": "replace",
            "scheme": "foo",
            "id": "B",
            "new_id": "D",
        },
    ]
    logger.log("abcde-12345", deltas[0])
    logger.log("abcde-12345", deltas[1])
    logger.flush()
    entries = logger.read()
    deltas = "A -> X + B -> D"
    assert deltas == entries[0]["deltas"]
    assert "" == entries[0]["error"]
